﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace FFT_Template
{
    class FFT : FourierDemo
    {
        public static Complex[][] ToComplex(float[][][] normalized_image, int channel)
        {
            Complex[][] result = new Complex[Size][];

            for (int x = 0; x < Size; x++)
            {
                result[x] = new Complex[Size];
                for (int y = 0; y < Size; y++)
                {
                    result[x][y] = new Complex(normalized_image[channel][x][y], 0);
                }
            }

            return result;
        }

        public static Complex[] Forward(Complex[] input, bool phaseShift = true)
        {
            var result = new Complex[input.Length];
            var omega = (float)(-2.0 * Math.PI / input.Length);

            if (input.Length == 1)
            {
                result[0] = input[0];

                if (Complex.IsNaN(result[0]))
                {
                    return new[] { new Complex(0, 0) };
                }
                return result;
            }

            var evenInput = new Complex[input.Length / 2];
            var oddInput = new Complex[input.Length / 2];

            for (int i = 0; i < input.Length / 2; i++)
            {
                evenInput[i] = input[2 * i];
                oddInput[i] = input[2 * i + 1];
            }

            var even = Forward(evenInput, phaseShift);
            var odd = Forward(oddInput, phaseShift);

            for (int k = 0; k < input.Length / 2; k++)
            {
                int phase;

                if (phaseShift)
                {
                    phase = k - Size / 2;
                }
                else
                {
                    phase = k;
                }
                odd[k] *= Complex.Polar(1, omega * phase);
            }

            for (int k = 0; k < input.Length / 2; k++)
            {
                result[k] = even[k] + odd[k];
                result[k + input.Length / 2] = even[k] - odd[k];
            }

            return result;
        }

        public static Complex[][] Forward(float[][][] normalized_image, int channel = 0)
        {
            var p = new Complex[Size][];
            var f = new Complex[Size][];
            var t = new Complex[Size][];

            var complex_image = ToComplex(normalized_image, channel);

            for (int l = 0; l < Size; l++)
            {
                p[l] = Forward(complex_image[l]);
            }

            for (int l = 0; l < Size; l++)
            {
                t[l] = new Complex[Size];
                for (int k = 0; k < Size; k++)
                {
                    t[l][k] = p[k][l];
                }
                f[l] = Forward(t[l]);
            }
            return f;
        }

        public static Complex[] Inverse(Complex[] input)
        {
            for (int i = 0; i < input.Length; i++)
            {
                input[i] = Complex.Conjugate(input[i]);
            }

            var transform = Forward(input, false);

            for (int i = 0; i < input.Length; i++)
            {
                transform[i] = Complex.Conjugate(transform[i]);
            }

            return transform;
        }

        public static float[][][] InverseToFloat(Complex[][][] frequencies)
        {
            float[][][] result = new float[3][][];

            for (int i = 0; i < 3; i++)
            {
                var p = new Complex[Size][];
                var f = new Complex[Size][];
                var t = new Complex[Size][];

                for (int l = 0; l < Size; l++)
                {
                    p[l] = Inverse(frequencies[i][l]);
                }

                for (int l = 0; l < Size; l++)
                {
                    t[l] = new Complex[Size];
                    for (int k = 0; k < Size; k++)
                    {
                        t[l][k] = p[k][l] / (Size * Size);
                    }
                    f[l] = Inverse(t[l]);
                }

                result[i] = new float[Size][];
                for (int x = 0; x < Size; x++)
                {
                    result[i][x] = new float[Size];
                    for (int y = 0; y < Size; y++)
                    {
                        result[i][x][y] = Complex.Modulus(f[x][y]);
                    }
                }
            }

            return result;
        }


    }
}
